from flask_monitoringdashboard.core.cache import update_duration_cache
from flask_monitoringdashboard.core.profiler.base_profiler import BaseProfiler
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.request import add_request


class PerformanceProfiler(BaseProfiler):
    """
    Used for updating the performance and utilization of the endpoint in the database.
    Used when monitoring-level == 1
    """

    def __init__(self, endpoint, ip, duration, group_by, status_code=200):
        super(PerformanceProfiler, self).__init__(endpoint)
        self._ip = ip
        self._duration = duration * 1000  # Conversion from sec to ms
        self._endpoint = endpoint
        self._group_by = group_by
        self._status_code = status_code

    def run(self):
        update_duration_cache(endpoint_name=self._endpoint.name, duration=self._duration)
        with session_scope() as session:
            add_request(
                session,
                duration=self._duration,
                endpoint_id=self._endpoint.id,
                ip=self._ip,
                group_by=self._group_by,
                status_code=self._status_code,
            )



from flask import jsonify

from flask_monitoringdashboard.controllers.profiler import get_profiler_table, get_grouped_profiler
from flask_monitoringdashboard.database import session_scope, Endpoint

from flask_monitoringdashboard.core.auth import secure

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.database.count import count_profiled_requests


@blueprint.route('/api/num_profiled/<endpoint_id>')
@secure
def num_profiled(endpoint_id):
    with session_scope() as session:
        return jsonify(count_profiled_requests(session, endpoint_id))


@blueprint.route('/api/profiler_table/<endpoint_id>/<offset>/<per_page>')
@secure
def profiler_table(endpoint_id, offset, per_page):
    with session_scope() as session:
        return jsonify(get_profiler_table(session, endpoint_id, offset, per_page))


@blueprint.route('/api/grouped_profiler/<endpoint_id>')
@secure
def grouped_profiler(endpoint_id):
    with session_scope() as session:
        return jsonify(get_grouped_profiler(session, endpoint_id))



import sys
import threading
import traceback

import psutil
from flask import request

from flask_monitoringdashboard import config
from flask_monitoringdashboard.core.cache import update_duration_cache, get_avg_endpoint
from flask_monitoringdashboard.core.logger import log
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.outlier import add_outlier
from flask_monitoringdashboard.database.request import add_request


class OutlierProfiler(threading.Thread):
    """
    Used for collecting additional information if the request is an outlier
    """

    def __init__(self, current_thread, endpoint, ip, group_by):
        threading.Thread.__init__(self)
        self._current_thread = current_thread
        self._endpoint = endpoint
        self._ip = ip
        self._group_by = group_by
        self._cpu_percent = None
        self._memory = None
        self._stacktrace = ''
        self._exit = threading.Event()

        self._request = str(request.headers), str(request.environ), request.url.encode('utf-8')

    def run(self):
        # sleep for average * ODC ms
        average = get_avg_endpoint(self._endpoint.name) * config.outlier_detection_constant
        self._exit.wait(average / 1000)
        if not self._exit.is_set():
            stack_list = []
            try:
                frame = sys._current_frames()[self._current_thread]
            except KeyError:
                log('Can\'t get the stacktrace of the main thread.')
                return
            in_endpoint_code = False
            # filename, line number, function name, source code line
            for fn, ln, fun, line in traceback.extract_stack(frame):
                if self._endpoint.name == fun:
                    in_endpoint_code = True
                if in_endpoint_code:
                    stack_list.append(
                        'File: "{}", line {}, in "{}": "{}"'.format(fn, ln, fun, line)
                    )

            # Set the values in the object
            self._stacktrace = '<br />'.join(stack_list)
            self._cpu_percent = str(psutil.cpu_percent(interval=None, percpu=True))
            self._memory = str(psutil.virtual_memory())

    def stop(self, duration, status_code):
        self._exit.set()
        update_duration_cache(endpoint_name=self._endpoint.name, duration=duration * 1000)
        with session_scope() as session:
            request_id = add_request(
                session,
                duration=duration * 1000,
                endpoint_id=self._endpoint.id,
                ip=self._ip,
                group_by=self._group_by,
                status_code=status_code,
            )
            if self._memory:
                add_outlier(
                    session,
                    request_id,
                    self._cpu_percent,
                    self._memory,
                    self._stacktrace,
                    self._request,
                )

    def stop_by_profiler(self):
        self._exit.set()

    def add_outlier(self, session, request_id):
        if self._memory:
            add_outlier(
                session,
                request_id,
                self._cpu_percent,
                self._memory,
                self._stacktrace,
                self._request,
            )



"""
Contains all functions that access a Request object.
"""
import time

from sqlalchemy import and_, func

from flask_monitoringdashboard.database import Request


def get_latencies_sample(session, endpoint_id, criterion, sample_size=500):
    query = (
        session.query(Request.duration).filter(Request.endpoint_id == endpoint_id,
                                               *criterion)
    )
    # return random rows: See https://stackoverflow.com/a/60815
    dialect = session.bind.dialect.name

    if dialect == 'sqlite':
        query = query.order_by(func.random())
    elif dialect == 'mysql':
        query = query.order_by(func.rand())

    query = query.limit(sample_size)

    return [item.duration for item in query.all()]


def add_request(session, duration, endpoint_id, ip, group_by, status_code):
    """ Adds a request to the database. Returns the id.
    :param status_code:  status code of the request
    :param session: session for the database
    :param duration: duration of the request
    :param endpoint_id: id of the endpoint
    :param ip: IP address of the requester
    :param group_by: a criteria by which the requests can be grouped
    :return the id of the request after it was stored in the database
    """
    request = Request(
        endpoint_id=endpoint_id,
        duration=duration,
        ip=ip,
        group_by=group_by,
        status_code=status_code,
    )
    session.add(request)
    session.commit()
    return request.id


def get_date_of_first_request(session):
    """ Returns the date (as unix timestamp) of the first request since FMD was deployed.
    :param session: session for the database
    :return time of the first request
    """
    result = session.query(Request.time_requested).order_by(
        Request.time_requested).first()
    if result:
        return int(time.mktime(result[0].timetuple()))
    return -1


def create_time_based_sample_criterion(start_date, end_date):
    return and_(Request.time_requested > start_date, Request.time_requested <= end_date)


def get_date_of_first_request_version(session, version):
    """ Returns the date (as unix timestamp) of the first request in the current FMD version.
    :param session: session for the database
    :param version: version of the dashboard
    :return time of the first request in that version
    """
    result = (
        session.query(Request.time_requested)
            .filter(Request.version_requested == version)
            .order_by(Request.time_requested)
            .first()
    )
    if result:
        return int(time.mktime(result[0].timetuple()))
    return -1



"""The app tracks the performance of various endpoints over time.
To bind, use the following lines of code:

>>> import flask_monitoringdashboard as dashboard
>>> from flask import Flask
    ...
>>> app = Flask(__name__)
    ...
>>> dashboard.bind(app)

The dashboard with the results that are collected can be found at:
    localhost:5000/dashboard
"""

import os

from flask import Blueprint

from flask_monitoringdashboard.core.config import Config
from flask_monitoringdashboard.core.logger import log


def loc():
    """Get the current location of the project."""
    return os.path.abspath(os.path.dirname(__file__)) + '/'


config = Config()
blueprint = Blueprint('dashboard', __name__, template_folder=loc() + 'templates')


def bind(app, schedule=True, include_dashboard=True):
    """Binding the app to this object should happen before importing the routing-
    methods below. Thus, the importing statement is part of this function.

    :param app: the app for which the performance has to be tracked
    :param schedule: flag telling if the background scheduler should be started
    :param include_dashboard: flag telling if the views should be added or not.
    """
    blueprint.name = config.blueprint_name
    config.app = app
    # Provide a secret-key for using WTF-forms
    if not app.secret_key:
        log('WARNING: You should provide a security key.')
        app.secret_key = 'my-secret-key'

    # Add all route-functions to the blueprint
    if include_dashboard:
        from flask_monitoringdashboard.views import (
            deployment,
            custom,
            endpoint,
            outlier,
            request,
            profiler,
            version,
            auth,
            reporting,
        )
        import flask_monitoringdashboard.views

    # Add wrappers to the endpoints that have to be monitored
    from flask_monitoringdashboard.core.measurement import init_measurement
    from flask_monitoringdashboard.core.cache import init_cache
    from flask_monitoringdashboard.core import custom_graph

    blueprint.record_once(lambda _state: init_measurement())
    blueprint.record_once(lambda _state: init_cache())
    if schedule:
        custom_graph.init(app)

    # register the blueprint to the app
    app.register_blueprint(blueprint, url_prefix='/' + config.link)

    # flush cache to db before shutdown
    import atexit
    from flask_monitoringdashboard.core.cache import flush_cache

    atexit.register(flush_cache)

    if not include_dashboard:
        @app.teardown_request
        def teardown(_):
            flush_cache()


def add_graph(title, func, trigger="interval", **schedule):
    """Add a custom graph to the dashboard. You must specify the following arguments:

    :param title: title of the graph (must be unique)
    :param schedule: dict containing values for weeks, days, hours, minutes, seconds
    :param func: function reference without arguments
    :param trigger: str|apscheduler.triggers.base.BaseTrigger trigger: trigger that determines when
            ``func`` is called
    """
    from flask_monitoringdashboard.core import custom_graph

    graph_id = custom_graph.register_graph(title)
    custom_graph.add_background_job(func, graph_id, trigger, **schedule)



import configparser
import os

from flask_monitoringdashboard.core.config.parser import (
    get_environment_var,
    parse_literal,
    parse_bool,
    parse_string,
    parse_version,
)


def test_init_from(config):
    """Test whether the group_by returns the right result."""

    config.init_from()
    config.init_from(file='../../config.cfg')


def test_parser():
    """Test whether the parser reads the right values."""

    parser = configparser.RawConfigParser()
    version = '1.2.3'
    string = 'string-value'
    bool = 'False'
    literal = "['a', 'b', 'c']"
    literal2 = '1.23'
    section = 'dashboard'

    parser.add_section(section)
    parser.set(section, 'APP_VERSION', version)
    parser.set(section, 'string', string)
    parser.set(section, 'bool', bool)
    parser.set(section, 'literal', literal)
    parser.set(section, 'literal2', literal2)

    assert parse_version(parser, section, 'default') == version
    assert parse_string(parser, section, 'string', 'default') == string
    assert not parse_bool(parser, section, 'bool', 'True')
    assert parse_literal(parser, section, 'literal', 'default') == ['a', 'b', 'c']
    assert parse_literal(parser, section, 'literal2', 'default') == 1.23




import datetime

from numpy import median
from sqlalchemy import and_

from flask_monitoringdashboard import config
from flask_monitoringdashboard.core import cache
from flask_monitoringdashboard.core.blueprints import get_blueprint
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.measurement import add_decorator
from flask_monitoringdashboard.core.timezone import to_local_datetime, to_utc_datetime
from flask_monitoringdashboard.core.utils import simplify
from flask_monitoringdashboard.database import Request
from flask_monitoringdashboard.database.count_group import count_requests_group, get_value
from flask_monitoringdashboard.database.data_grouped import (
    get_endpoint_data_grouped,
    get_user_data_grouped,
    get_version_data_grouped,
)
from flask_monitoringdashboard.database.endpoint import (
    get_last_requested,
    get_endpoints,
    get_endpoint_by_name,
    update_endpoint,
)
from flask_monitoringdashboard.database.versions import get_first_requests


def get_endpoint_overview(session):
    """
    :param session: session for the database
    :return: A list of properties for each endpoint that is found in the database
    """
    week_ago = datetime.datetime.utcnow() - datetime.timedelta(days=7)
    now_local = to_local_datetime(datetime.datetime.utcnow())
    today_local = now_local.replace(hour=0, minute=0, second=0, microsecond=0)
    today_utc = to_utc_datetime(today_local)

    # First flush last requested info to db
    cache.flush_cache()
    error_hits_criterion = and_(Request.status_code >= 400, Request.status_code < 600)

    hits_today = count_requests_group(session, Request.time_requested > today_utc)
    hits_today_errors = count_requests_group(
        session, and_(Request.time_requested > today_utc, error_hits_criterion)
    )

    hits_week = count_requests_group(session, Request.time_requested > week_ago)
    hits_week_errors = count_requests_group(
        session, and_(Request.time_requested > week_ago, error_hits_criterion)
    )

    hits = count_requests_group(session)

    median_today = get_endpoint_data_grouped(session, median, Request.time_requested > today_utc)
    median_week = get_endpoint_data_grouped(session, median, Request.time_requested > week_ago)
    median_overall = get_endpoint_data_grouped(session, median)
    access_times = get_last_requested(session)

    return [
        {
            'id': endpoint.id,
            'name': endpoint.name,
            'blueprint': get_blueprint(endpoint.name),
            'monitor': endpoint.monitor_level,
            'color': get_color(endpoint.name),
            'hits-today': get_value(hits_today, endpoint.id),
            'hits-today-errors': get_value(hits_today_errors, endpoint.id),
            'hits-week': get_value(hits_week, endpoint.id),
            'hits-week-errors': get_value(hits_week_errors, endpoint.id),
            'hits-overall': get_value(hits, endpoint.id),
            'median-today': get_value(median_today, endpoint.id),
            'median-week': get_value(median_week, endpoint.id),
            'median-overall': get_value(median_overall, endpoint.id),
            'last-accessed': get_value(access_times, endpoint.name, default=None),
        }
        for endpoint in get_endpoints(session)
    ]


def get_endpoint_users(session, endpoint_id, users):
    """
    :param session: session for the database
    :param endpoint_id: id for the endpoint
    :param users: a list of users to be filtered on
    :return: a list of dicts with the performance of each user
    """
    times = get_user_data_grouped(
        session, lambda x: simplify(x, 100), Request.endpoint_id == endpoint_id
    )
    first_requests = get_first_requests(session, endpoint_id)
    return [
        {
            'user': u,
            'date': get_value(first_requests, u),
            'values': get_value(times, u),
            'color': get_color(u),
        }
        for u in users
    ]


def get_endpoint_versions(session, endpoint_id, versions):
    """
    :param session: session for the database
    :param endpoint_id: id for the endpoint
    :param versions: a list of version to be filtered on
    :return: a list of dicts with the performance of each version
    """
    times = get_version_data_grouped(
        session, lambda x: simplify(x, 100), Request.endpoint_id == endpoint_id
    )
    first_requests = get_first_requests(session, endpoint_id)
    return [
        {
            'version': v,
            'date': get_value(first_requests, v),
            'values': get_value(times, v),
            'color': get_color(v),
        }
        for v in versions
    ]


def get_api_performance(session, endpoints):
    """
    :param session: session for the database
    :param endpoints: a list of endpoints, encoded by their name
    :return: for every endpoint in endpoints, a list with the performance
    """
    db_endpoints = [get_endpoint_by_name(session, end) for end in endpoints]
    data = get_endpoint_data_grouped(session, lambda x: simplify(x, 10))
    return [
        {'name': end.name, 'values': get_value(data, end.id, default=[])}
        for end in db_endpoints
    ]


def set_endpoint_rule(session, endpoint_name, monitor_level):
    """
    :param session: session for the database
    :param endpoint_name: name of the endpoint
    :param monitor_level: integer, representing the monitoring-level
    """
    update_endpoint(session, endpoint_name, value=monitor_level)

    # Remove wrapper
    original = getattr(config.app.view_functions[endpoint_name], 'original', None)
    if original:
        config.app.view_functions[endpoint_name] = original
    session.commit()

    add_decorator(get_endpoint_by_name(session, endpoint_name))



"""
    Use this file for migrating your data from SQLite to MySQL. It only works for version >=2.0.0
    of the Flask Monitoring Dashboard. For migrating your data from version 1.x.x to 2.y.y, use
    the script "migrate_v1_to_v2.py".

    Before running the script, make sure to change the SQLITE_URL and MYSQL_URL variables. Refer to
    http://docs.sqlalchemy.org/en/latest/core/engines.html on how to configure them. Also, make sure
    MySQL Server is running on your machine.
"""
import timeit

from contextlib import contextmanager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from flask_monitoringdashboard.database import Base, Endpoint, Request, Outlier, CodeLine, StackLine

SQLITE_URL = 'sqlite:////Users/user/Flask-MonitoringDashboard/fmd.db'
MYSQL_URL = 'mysql+pymysql://user:password@localhost:3306/database'
ENTITIES = [Endpoint, Request, Outlier, CodeLine, StackLine]
CHUNK_SIZE = 10000


def create_db_session(db_url, clean=False):
    engine = create_engine(db_url)
    if clean:
        Base.metadata.drop_all(engine)
        Base.metadata.create_all(engine)
    Base.metadata.bind = engine
    return sessionmaker(bind=engine)


@contextmanager
def session_scope(session):
    session = session()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def get_row_count(db_session_old, entity_class):
    with session_scope(db_session_old) as session:
        count = session.query(entity_class).count()
        session.expunge_all()
    return count


def get_old_instances(db_session_old, entity_class, start, end):
    with session_scope(db_session_old) as session:
        old_instances = session.query(entity_class).slice(start, end).all()
        session.expunge_all()
    return old_instances


def migrate_batch(old_instances, db_session_new, entity_class):
    new_instances = []
    with session_scope(db_session_new) as session:
        for (index, old_instance) in enumerate(old_instances):
            # _sa_instance_state is a non-db value used internally by SQLAlchemy
            del old_instance.__dict__['_sa_instance_state']
            new_instances.append(entity_class(**old_instance.__dict__))
        session.bulk_save_objects(new_instances)


def migrate_all(db_session_old, db_session_new, entity_class, count):
    chunks = count // CHUNK_SIZE
    for i in range(chunks + 1):
        print("Moving batch %d of %d for entity %s..." % (i + 1, chunks + 1, entity_class.__name__))
        old_instances = get_old_instances(
            db_session_old, entity_class, i * CHUNK_SIZE, (i + 1) * CHUNK_SIZE
        )
        migrate_batch(old_instances, db_session_new, entity_class)


def main():
    db_session_old = create_db_session(SQLITE_URL)
    db_session_new = create_db_session(MYSQL_URL, clean=True)
    t0 = timeit.default_timer()
    t_previous = t0
    for entity_class in ENTITIES:
        count = get_row_count(db_session_old, entity_class)
        migrate_all(db_session_old, db_session_new, entity_class, count)
        t_now = timeit.default_timer()
        print("Moving %s took %f seconds" % (entity_class.__name__, t_now - t_previous))
        t_previous = t_now
    print("Total time was %f seconds" % (t_previous - t0))


if __name__ == "__main__":
    main()



import threading

from flask_monitoringdashboard.core.cache import update_last_requested_cache


class BaseProfiler(threading.Thread):
    """
    Only updates the last_accessed time in the database for a certain endpoint.
    Used for monitoring-level == 0
    """

    def __init__(self, endpoint):
        self._endpoint = endpoint
        threading.Thread.__init__(self)

    def run(self):
        update_last_requested_cache(endpoint_name=self._endpoint.name)



from sqlalchemy import func, distinct

from flask_monitoringdashboard.database import Request, StackLine


def count_rows(session, column, *criterion):
    """
    Count the number of rows of a specified column.
    :param session: session for the database
    :param column: column to count
    :param criterion: where-clause of the query
    :return: number of rows
    """
    return session.query(func.count(distinct(column))).filter(*criterion).scalar()


def count_requests(session, endpoint_id, *where):
    """
    Return the number of hits for a specific endpoint (possible with more filter arguments).
    :param session: session for the database
    :param endpoint_id: id of the endpoint
    :param where: additional arguments
    """
    return count_rows(session, Request.id, Request.endpoint_id == endpoint_id, *where)


def count_total_requests(session, *where):
    """
    Return the number of total hits
    :param session: session for the database
    :param where: additional arguments
    """
    return count_rows(session, Request.id, *where)


def count_outliers(session, endpoint_id):
    """
    :param session: session for the database
    :param endpoint_id: id of the endpoint
    :return: An integer with the number of rows in the Outlier-table.
    """
    return count_rows(session, Request.id, Request.endpoint_id == endpoint_id, Request.outlier)


def count_profiled_requests(session, endpoint_id):
    """
    Count the number of profiled requests for a certain endpoint
    :param session: session for the database
    :param endpoint_id: id of the endpoint
    :return: An integer
    """
    return (
        session.query(func.count(distinct(StackLine.request_id)))
        .filter(Request.endpoint_id == endpoint_id)
        .join(Request.stack_lines)
        .scalar()
    )



import pytest


@pytest.fixture
def filename():
    return 'filename0'


@pytest.fixture
def line_number():
    return 42


def test_get_path(path_hash, filename, line_number):
    assert path_hash.get_path(filename, line_number) == '0:42'


def test_append(path_hash, filename, line_number):
    path_hash.get_path(filename, line_number)
    assert path_hash.append(filename, line_number) == '0:42->0:42'


def test_encode(path_hash, filename, line_number):
    assert path_hash._encode(filename, line_number) == '0:42'


def test_decode(path_hash, filename, line_number):
    assert path_hash._decode(path_hash._encode(filename, line_number)) == (filename, line_number)


def test_get_indent(path_hash, filename, line_number):
    assert path_hash.get_indent('') == 0
    assert path_hash.get_indent('0:42') == 1
    assert path_hash.get_indent('0:42->0:42') == 2


def test_get_last_fn_ln(path_hash, filename, line_number):
    assert path_hash.get_last_fn_ln(path_hash._encode(filename, line_number)) == (filename, line_number)


def test_get_code(path_hash, stack_line, filename, line_number):
    path = path_hash.get_stacklines_path([stack_line], 0)
    assert path_hash.get_code(path) == stack_line.code.code


def test_get_stacklines_path(path_hash, stack_line, stack_line_2, filename, line_number):
    assert path_hash.get_stacklines_path([stack_line, stack_line_2], 0) == '1:0'
    assert path_hash.get_stacklines_path([stack_line, stack_line_2], 1) == '1:0->1:0'



import pytest

from flask_monitoringdashboard.core.measurement import init_measurement, add_decorator


@pytest.mark.usefixtures('request_context')
def test_init_measurement():
    init_measurement()


@pytest.mark.usefixtures('request_context')
@pytest.mark.parametrize('endpoint__monitor_level', [0, 1, 2, 3])
def test_add_decorator(endpoint, config):
    def f():
        pass

    config.app.view_functions[endpoint.name] = f
    add_decorator(endpoint)
    assert config.app.view_functions[endpoint.name].original == f


@pytest.mark.usefixtures('request_context')
@pytest.mark.parametrize('endpoint__monitor_level', [-1])
def test_add_decorator_fails(endpoint, config):
    config.app.view_functions[endpoint.name] = lambda: 42
    with pytest.raises(ValueError):
        endpoint.monitor_level = -1
        add_decorator(endpoint)



from flask_monitoringdashboard.database import User
from flask_monitoringdashboard.database.auth import get_user


def test_get_user_adds_default(session, config):
    session.query(User).delete()  # delete all existing users
    session.commit()
    new_user = get_user(config.username, config.password)

    assert session.query(User).count() == 1

    assert new_user.username == config.username
    assert new_user.check_password(config.password)


def test_get_user_returns_none(user):
    """Test that get_user returns None if the user cannot be found."""
    assert get_user(username=user.username, password='1234') is None



"""
    Contains all functions that are used to track the performance of the flask-application.
    See init_measurement() for more detailed info.
"""
import time
from functools import wraps

from flask_monitoringdashboard import config
from flask_monitoringdashboard.core.profiler import (
    start_thread_last_requested,
    start_performance_thread,
    start_outlier_thread,
    start_profiler_and_outlier_thread,
)
from flask_monitoringdashboard.core.rules import get_rules
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.endpoint import get_endpoint_by_name


def init_measurement():
    """
    This should be added to the list of functions that are executed before the first request.
    This function is used in the config-method in __init__ of this folder
    It adds wrappers to the endpoints for tracking their performance and last access times.
    """
    with session_scope() as session:
        for rule in get_rules():
            endpoint = get_endpoint_by_name(session, rule.endpoint)
            add_decorator(endpoint)


def add_decorator(endpoint):
    """
    Adds a wrapper to an endpoint in the app, depending on the monitor level.
    :param endpoint: Endpoint object to be monitored
    """
    fun = config.app.view_functions[endpoint.name]
    if endpoint.monitor_level == 0:
        add_wrapper0(endpoint, fun)
    elif endpoint.monitor_level == 1:
        add_wrapper1(endpoint, fun)
    elif endpoint.monitor_level == 2:
        add_wrapper2(endpoint, fun)
    elif endpoint.monitor_level == 3:
        add_wrapper3(endpoint, fun)
    else:
        raise ValueError('Incorrect monitoringLevel')


def add_wrapper0(endpoint, fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        result = fun(*args, **kwargs)
        start_thread_last_requested(endpoint)
        return result

    wrapper.original = fun
    config.app.view_functions[endpoint.name] = wrapper


def is_valid_status_code(status_code):
    """
    Returns whether the input is a valid status code. A status code is valid if it's an integer value and in the
    range [100, 599] :param status_code: :return:
    """
    return type(status_code) == int and 100 <= status_code < 600


def status_code_from_response(result):
    """
    Extracts the status code from the result that was returned from the route handler.

    :param result: The return value of the route handler
    :return:
    """
    if type(result) == str:
        return 200

    status_code = 200  # default

    # Pull it from a tuple
    if isinstance(result, tuple):
        status_code = result[1]
    else:
        # Try to pull it from an object
        try:
            status_code = getattr(result, 'status_code')
        except:
            pass

    if not is_valid_status_code(status_code):
        return 500

    return status_code


def evaluate(route_handler, args, kwargs):
    """
    Invokes the given route handler and extracts the return value, status_code and the exception if it was raised

    :param route_handler:
    :param args:
    :param kwargs:
    :return:
    """
    try:
        result = route_handler(*args, **kwargs)
        status_code = status_code_from_response(result)

        return result, status_code, None

    except Exception as e:
        return None, 500, e


def add_wrapper1(endpoint, fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        start_time = time.time()

        result, status_code, raised_exception = evaluate(fun, args, kwargs)

        duration = time.time() - start_time
        start_performance_thread(endpoint, duration, status_code)

        if raised_exception:
            raise raised_exception

        return result

    wrapper.original = fun
    config.app.view_functions[endpoint.name] = wrapper


def add_wrapper2(endpoint, fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        outlier = start_outlier_thread(endpoint)
        start_time = time.time()

        result, status_code, raised_exception = evaluate(fun, args, kwargs)

        duration = time.time() - start_time
        outlier.stop(duration, status_code)

        if raised_exception:
            raise raised_exception

        return result

    wrapper.original = fun
    config.app.view_functions[endpoint.name] = wrapper


def add_wrapper3(endpoint, fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        thread = start_profiler_and_outlier_thread(endpoint)
        start_time = time.time()

        result, status_code, raised_exception = evaluate(fun, args, kwargs)

        duration = time.time() - start_time
        thread.stop(duration, status_code)

        if raised_exception:
            raise raised_exception

        return result

    wrapper.original = fun
    config.app.view_functions[endpoint.name] = wrapper



"""
    Helper functions for parsing the arguments from the config file
"""
import ast
import os

from flask_monitoringdashboard.core.logger import log


def parse_version(parser, header, version):
    """
    Parse the version given in the config-file.
    If both GIT and VERSION are used, the GIT argument is used.
    :param parser: the parser to be used for parsing
    :param header: name of the header in the configuration file
    :param version: the default version
    """
    version = parse_string(parser, header, 'APP_VERSION', version)
    if parser.has_option(header, 'GIT'):
        git = parser.get(header, 'GIT')
        try:
            # current hash can be found in the link in HEAD-file in git-folder
            # The file is specified by: 'ref: <location>'
            git_head = os.path.join(git, 'HEAD')
            if os.path.isfile(git_head):
                git_file = (open(git_head).read().rsplit(': ', 1)[1]).rstrip()
                # read the git-version
                version_file = os.path.join(git, git_file)
                if os.path.exists(version_file):
                    version = open(version_file).read()
                    # cut version to at most 6 chars
                    return version[:6]
            else:
                # Return "dummy" version in case of no git version file found
                log("Folder {} not found, using dummy version: {}".format(git_head, version))
                return version
        except IOError:
            log("Error reading one of the files to retrieve the current git-version.")
            raise
    return version


def parse_string(parser, header, arg_name, arg_value):
    """
    Parse an argument from the given parser. If the argument is not specified, return the default
    value
    :param parser: the parser to be used for parsing
    :param header: name of the header in the configuration file
    :param arg_name: name in the configuration file
    :param arg_value: default value, the the value is not found
    """
    env = get_environment_var(arg_name)
    arg_value = env if env else arg_value
    if parser.has_option(header, arg_name):
        return parser.get(header, arg_name)
    return arg_value


def parse_bool(parser, header, arg_name, arg_value):
    """
    Parse an argument from the given parser. If the argument is not specified, return the default
    value
    :param parser: the parser to be used for parsing
    :param header: name of the header in the configuration file
    :param arg_name: name in the configuration file
    :param arg_value: default value, the the value is not found
    """
    env = get_environment_var(arg_name)
    arg_value = env if env else arg_value
    if parser.has_option(header, arg_name):
        return parser.get(header, arg_name) == 'True'
    return arg_value


def parse_literal(parser, header, arg_name, arg_value):
    """
    Parse an argument from the given parser. If the argument is not specified, return the default
    value
    :param parser: the parser to be used for parsing
    :param header: name of the header in the configuration file
    :param arg_name: name in the configuration file
    :param arg_value: default value, the the value is not found
    """
    env = get_environment_var(arg_name)
    arg_value = ast.literal_eval(env) if env else arg_value
    if parser.has_option(header, arg_name):
        return ast.literal_eval(parser.get(header, arg_name))
    return arg_value


def get_environment_var(environment_var):
    """
    Retrieve the arg_value from the environment variable
    :param environment_var: name of the environment variable
    :return: either the value of the environment_var or None
    """
    return os.environ.get(environment_var, None)



import datetime

from flask.json import jsonify

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.custom_graph import get_custom_graphs
from flask_monitoringdashboard.database import row2dict, session_scope
from flask_monitoringdashboard.database.custom_graph import get_graph_data


@blueprint.route('/api/custom_graphs')
@secure
def custom_graphs():
    graphs = get_custom_graphs()
    if not graphs:
        return jsonify([])
    return jsonify([row2dict(row) for row in graphs if row is not None])


@blueprint.route('/api/custom_graph/<graph_id>/<start_date>/<end_date>')
@secure
def custom_graph(graph_id, start_date, end_date):
    start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d')
    end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d')
    with session_scope() as session:
        return jsonify(get_graph_data(session, graph_id, start_date, end_date))



import datetime

from flask import jsonify

from flask_monitoringdashboard import blueprint, config
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.timezone import to_local_datetime
from flask_monitoringdashboard.core.utils import get_details
from flask_monitoringdashboard.database import session_scope


@blueprint.route('/api/deploy_details')
@secure
def deploy_details():
    """
    :return: A JSON-object with deployment details
    """
    with session_scope() as session:
        details = get_details(session)
    details['first-request'] = to_local_datetime(
        datetime.datetime.fromtimestamp(details['first-request'])
    )
    details['first-request-version'] = to_local_datetime(
        datetime.datetime.fromtimestamp(details['first-request-version'])
    )
    return jsonify(details)


@blueprint.route('/api/deploy_config')
@secure
def deploy_config():
    """
    :return: A JSON-object with configuration details
    """
    return jsonify(
        {
            'database_name': config.database_name,
            'outlier_detection_constant': config.outlier_detection_constant,
            'timezone': str(config.timezone),
            'colors': config.colors,
        }
    )



from flask_monitoringdashboard.database import CodeLine


def get_code_line(session, fn, ln, name, code):
    """
    Get a CodeLine object from a given quadruple of fn, ln, name, code. If the CodeLine object
    doesn't already exist, a new one is created in the database.
    :param session: session for the database
    :param fn: filename (string)
    :param ln: line_number of the code (int)
    :param name: function name (string)
    :param code: line of code (string)
    :return: a CodeLine object
    """
    result = (
        session.query(CodeLine)
        .filter(
            CodeLine.filename == fn,
            CodeLine.line_number == ln,
            CodeLine.function_name == name,
            CodeLine.code == code,
        )
        .first()
    )
    if not result:
        result = CodeLine(filename=fn, line_number=ln, function_name=name, code=code)
        session.add(result)
        session.flush()

    return result



import datetime


def to_local_datetime(dt):
    """
    Convert datetime (UTC) to local datetime based on the configuration.
    :param dt: UTC datetime object
    :return local datetime
    """
    from flask_monitoringdashboard import config

    if dt:
        return dt + config.timezone.utcoffset(datetime.datetime.utcnow())
    return None


def to_utc_datetime(dt):
    """
    Convert datetime (local) to UTC datetime based on the configuration.
    :param dt: local datetime object
    :return UTC datetime
    """
    from flask_monitoringdashboard import config

    if dt:
        return dt - config.timezone.utcoffset(datetime.datetime.utcnow())
    return None



from flask_monitoringdashboard.database import User, session_scope
from flask_monitoringdashboard import config


def get_user(username, password):
    """Validates the username and password and returns an User-object if both are valid.
    In case the User-table is empty, a user with default credentials is added.
    """
    with session_scope() as session:
        if session.query(User).count() == 0:
            user = User(username=config.username, is_admin=True)
            user.set_password(password=config.password)
            session.add(user)

        user = session.query(User).filter(User.username == username).one_or_none()
        if user is not None:
            if user.check_password(password=password):
                session.expunge_all()
                return user

    return None


def get_all_users(session):
    users = session.query(User).order_by(User.id).all()

    return [
        {
            'id': user.id,
            'username': user.username,
            'is_admin': user.is_admin,
        }
        for user in users
    ]



import ast

from flask_monitoringdashboard.core.logger import log
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.timezone import to_local_datetime
from flask_monitoringdashboard.core.utils import simplify
from flask_monitoringdashboard.database import row2dict
from flask_monitoringdashboard.database.outlier import get_outliers_cpus, get_outliers_sorted


def get_outlier_graph(session, endpoint_id):
    """
    :param session: session for the database
    :param endpoint_id: id of the endpoint
    :return: a list with data about each CPU performance
    """
    all_cpus = get_outliers_cpus(session, endpoint_id)
    cpu_data = [ast.literal_eval(cpu) for cpu in all_cpus]

    return [
        {'name': 'CPU core %d' % idx, 'values': simplify(data, 50), 'color': get_color(idx)}
        for idx, data in enumerate(zip(*cpu_data))
    ]


def get_outlier_table(session, endpoint_id, offset, per_page):
    """
    :param session: session for the database
    :param endpoint_id: id of the endpoint
    :param offset: number of items to be skipped
    :param per_page: maximum number of items to be returned
    :return: a list of length at most 'per_page' with data about each outlier
    """
    table = get_outliers_sorted(session, endpoint_id, offset, per_page)
    for idx, row in enumerate(table):
        row.request.time_requested = to_local_datetime(row.request.time_requested)
        try:
            row.request_url = row.request_url.decode('utf-8')
        except Exception as e:
            log(e)
        dict_request = row2dict(row.request)
        table[idx] = row2dict(row)
        table[idx]['request'] = dict_request
    return table



from flask_monitoringdashboard.controllers.requests import \
    get_status_code_frequencies_in_interval
from flask_monitoringdashboard.core.reporting.questions.report_question import (
    ReportAnswer,
    ReportQuestion,
)
from flask_monitoringdashboard.database import session_scope


class StatusCodeDistributionReportAnswer(ReportAnswer):
    def __init__(self, is_significant=False, percentages=None):
        super().__init__('STATUS_CODE_DISTRIBUTION')

        self.percentages = percentages
        self._is_significant = is_significant

    def is_significant(self):
        return self._is_significant

    def meta(self):
        return dict(percentages=self.percentages)


def frequency_to_percentage(freq, total):
    if total == 0:
        raise ValueError('`total` can not be zero!')

    return (float(freq)) / total * 100


class StatusCodeDistribution(ReportQuestion):
    MIN_NUM_REQUESTS = 30
    MIN_PERCENTAGE_DIFF_THRESHOLD = 3

    def get_answer(self, endpoint, requests_criterion, baseline_requests_criterion):

        with session_scope() as session:

            frequencies = get_status_code_frequencies_in_interval(session, endpoint.id,
                                                                  requests_criterion)

            baseline_frequencies = get_status_code_frequencies_in_interval(
                session,
                endpoint.id,
                baseline_requests_criterion
            )

        # all monitored status codes in both intervals
        all_monitored_status_codes = set(baseline_frequencies.keys()).union(
            set(frequencies.keys()))

        total_requests = sum(frequencies.values())
        total_baseline_requests = sum(baseline_frequencies.values())

        if (
            total_requests < self.MIN_NUM_REQUESTS
            or total_baseline_requests < self.MIN_NUM_REQUESTS
        ):
            return StatusCodeDistributionReportAnswer(is_significant=False)

        percentages = []
        max_percentage_diff = 0

        for status_code in all_monitored_status_codes:
            count = frequencies[status_code] if status_code in frequencies else 0

            baseline_count = (
                baseline_frequencies[status_code] if status_code in baseline_frequencies else 0
            )

            percentage = frequency_to_percentage(freq=count, total=total_requests)
            baseline_percentage = frequency_to_percentage(
                freq=baseline_count, total=total_baseline_requests
            )

            percentage_diff = percentage - baseline_percentage

            percentages.append(
                dict(
                    status_code=status_code,
                    comparison=percentage,
                    baseline=baseline_percentage,
                    percentage_diff=percentage_diff,
                )
            )

            max_percentage_diff = max(max_percentage_diff, percentage_diff)

        return StatusCodeDistributionReportAnswer(
            is_significant=max_percentage_diff > self.MIN_PERCENTAGE_DIFF_THRESHOLD,
            percentages=percentages,
        )



import pytest


@pytest.mark.usefixtures('outlier_1', 'outlier_2')
def test_num_outliers(dashboard_user, endpoint):
    response = dashboard_user.get('dashboard/api/num_outliers/{0}'.format(endpoint.id))
    assert response.status_code == 200

    data = response.json
    assert data == 2


@pytest.mark.parametrize('outlier_1__cpu_percent', ['[0, 1, 2, 3]'])
@pytest.mark.usefixtures('outlier_1')
def test_outlier_graph(dashboard_user, endpoint):
    response = dashboard_user.get('dashboard/api/outlier_graph/{0}'.format(endpoint.id))
    assert response.status_code == 200

    data = response.json
    for i in range(4):
        assert data[i]['name'] == "CPU core {0}".format(i)
        assert data[i]['values'] == [i]


@pytest.mark.parametrize('outlier_1__cpu_percent', ['[0, 1, 2, 3]'])
@pytest.mark.parametrize('outlier_1__memory', ['memory'])
@pytest.mark.parametrize('outlier_1__request_environment', ['request_environment'])
@pytest.mark.parametrize('outlier_1__request_header', ['request_header'])
@pytest.mark.parametrize('outlier_1__request_url', ['request_url'])
@pytest.mark.parametrize('outlier_1__stacktrace', ['stacktrace'])
@pytest.mark.parametrize('offset,per_page', [[0, 10]])
def test_outlier_table(dashboard_user, outlier_1, endpoint, offset, per_page):
    response = dashboard_user.get(
        'dashboard/api/outlier_table/{0}/{1}/{2}'.format(endpoint.id, offset, per_page),
    )
    assert response.status_code == 200

    [data] = response.json
    assert data['cpu_percent'] == outlier_1.cpu_percent
    assert data['id'] == str(outlier_1.id)
    assert data['memory'] == outlier_1.memory
    assert data['request_id'] == str(outlier_1.request.id)
    assert data['request_environment'] == outlier_1.request_environment
    assert data['request_header'] == outlier_1.request_header
    assert data['request_url'] == outlier_1.request_url
    assert data['stacktrace'] == outlier_1.stacktrace



from flask import jsonify, request, json

from flask_monitoringdashboard.controllers.versions import (
    get_multi_version_data,
    get_version_user_data,
    get_version_ip_data,
)
from flask_monitoringdashboard.database import session_scope

from flask_monitoringdashboard.core.auth import secure

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.database.versions import get_versions


@blueprint.route('/api/versions')
@blueprint.route('/api/versions/<endpoint_id>')
@secure
def versions(endpoint_id=None):
    """
    :param endpoint_id: integer
    :return: A JSON-list with all versions of a specific endpoint (version represented by a string)
    """
    with session_scope() as session:
        version_dates = get_versions(session, endpoint_id)
        dicts = []
        for vt in version_dates:
            dicts.append({'version': vt[0], 'date': vt[1]})
        return jsonify(dicts)


@blueprint.route('/api/multi_version', methods=['POST'])
@secure
def multi_version():
    """
        input must be a JSON-object, with a list of endpoints and versions, such as:
          {
            endpoints: ['endpoint0', endpoint1],
            versions: ['0.1', '0.2', '0.3']
          }
    :return: A JSON-list for all endpoints, with a JSON-list for every version.
        output: {
            [
              [10, 11, 12],
              [13, 14, 15]
            ]
          }
    """
    data = json.loads(request.data)['data']
    endpoints = data['endpoints']
    versions = data['versions']
    with session_scope() as session:
        return jsonify(get_multi_version_data(session, endpoints, versions))


@blueprint.route('/api/version_user/<endpoint_id>', methods=['POST'])
@secure
def version_user(endpoint_id):
    """
        input must be a JSON-object, with a list of versions and users, such as:
          {
            users: ['user0', user1],
            versions: ['0.1', '0.2', '0.3']
          }
    :return: A JSON-list for all users, with a JSON-list for every version.
        output: {
            data: [
              [10, 11, 12],
              [13, 14, 15]
            ],
            versions: [
              { date: '...', version: '0.1'},
              { date: '...', version: '0.2'},
              { date: '...', version: '0.3'}
            ]
          }
    """
    data = json.loads(request.data)['data']
    versions = data['versions']
    users = data['users']

    with session_scope() as session:
        return jsonify(get_version_user_data(session, endpoint_id, versions, users))


@blueprint.route('/api/version_ip/<endpoint_id>', methods=['POST'])
@secure
def version_ip(endpoint_id):
    """
        input must be a JSON-object, with a list of versions and IP-addresses, such as:
          {
            ip: ['127.0.0.1', '127.0.0.2'],
            versions: ['0.1', '0.2', '0.3']
          }
    :return: A JSON-list for all IP-addresses, with a JSON-list for every version.
        output: {
            data: [
              [10, 11, 12],
              [13, 14, 15]
            ],
            versions: [
              { date: '...', version: '0.1'},
              { date: '...', version: '0.2'},
              { date: '...', version: '0.3'}
            ]
          }
    """
    data = json.loads(request.data)['data']
    versions = data['versions']
    ips = data['ip']

    with session_scope() as session:
        return jsonify(get_version_ip_data(session, endpoint_id, versions, ips))



"""
This file contains all unit tests for the monitor-rules-table in the database. (Corresponding
to the file: 'flask_monitoringdashboard/database/outlier.py')
"""

import pytest

from flask_monitoringdashboard.database import Outlier
from flask_monitoringdashboard.database.count import count_outliers
from flask_monitoringdashboard.database.outlier import add_outlier, get_outliers_sorted, get_outliers_cpus


def test_add_outlier(session, request_1):
    assert not request_1.outlier

    add_outlier(
        session,
        request_id=request_1.id,
        cpu_percent="cpu_percent",
        memory="memory",
        stacktrace="stacktrace",
        request=("headers", "environ", "url"),
    )
    session.commit()
    assert session.query(Outlier).filter(Outlier.request_id == request_1.id).one()


def test_get_outliers(session, outlier_1, endpoint):
    outliers = get_outliers_sorted(session, endpoint_id=endpoint.id, offset=0, per_page=10)
    assert len(outliers) == 1
    assert outliers[0].id == outlier_1.id


@pytest.mark.usefixtures('outlier_1', 'outlier_2')
def test_count_outliers(session, endpoint):
    assert count_outliers(session, endpoint.id) == 2


@pytest.mark.usefixtures('outlier_1', 'outlier_2')
@pytest.mark.parametrize('outlier_1__cpu_percent', ['[0, 1, 2, 3]'])
@pytest.mark.parametrize('outlier_2__cpu_percent', ['[1, 2, 3, 4]'])
def test_get_outliers_cpus(session, endpoint):
    expected_cpus = ['[{0}, {1}, {2}, {3}]'.format(i, i + 1, i + 2, i + 3) for i in range(2)]
    assert get_outliers_cpus(session, endpoint.id) == expected_cpus



from flask import jsonify, request, json

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.controllers.endpoints import (
    get_endpoint_overview,
    get_api_performance,
    set_endpoint_rule,
    get_endpoint_versions,
    get_endpoint_users,
)
from flask_monitoringdashboard.controllers.requests import (
    get_status_code_distribution,
    get_error_requests,
)
from flask_monitoringdashboard.core.auth import secure, admin_secure
from flask_monitoringdashboard.core.utils import get_endpoint_details
from flask_monitoringdashboard.database import session_scope, row2dict
from flask_monitoringdashboard.database.endpoint import (
    get_users,
    get_ips,
    get_endpoints,
    get_endpoints_hits,
)


@blueprint.route('/api/overview')
@secure
def get_overview():
    """
    Get information per endpoint about the number of hits and median execution time
    :return: A JSON-list with a JSON-object per endpoint
    """
    with session_scope() as session:
        return jsonify(get_endpoint_overview(session))


@blueprint.route('/api/users/<endpoint_id>')
@secure
def users(endpoint_id):
    """
    :param endpoint_id: integer
    :return: A JSON-list with all users of a specific endpoint (user represented by a string)
    """
    with session_scope() as session:
        users_hits = get_users(session, endpoint_id)
        dicts = []
        for uh in users_hits:
            dicts.append({'user': uh[0], 'hits': uh[1]})
        return jsonify(dicts)


@blueprint.route('/api/ip/<endpoint_id>')
@secure
def ips(endpoint_id):
    """
    :param endpoint_id: integer
    :return: A JSON-list with all IP-addresses of a specific endpoint (ip represented by a string)
    """
    with session_scope() as session:
        ips_hits = get_ips(session, endpoint_id)
        dicts = []
        for ih in ips_hits:
            dicts.append({'ip': ih[0], 'hits': ih[1]})
        return jsonify(dicts)


@blueprint.route('/api/endpoints')
@secure
def endpoints():
    """
    :return: A JSON-list with information about every endpoint (encoded in a JSON-object)
        For more information per endpoint, see :func: get_overview
    """
    with session_scope() as session:
        return jsonify([row2dict(row) for row in get_endpoints(session)])


@blueprint.route('/api/endpoints_hits')
@secure
def endpoints_hits():
    """
    :return: A JSON-list with information about every endpoint and its total number of hits
    (encoded in a JSON-object)
        For more information per endpoint, see :func: get_overview
    """
    with session_scope() as session:
        end_hits = get_endpoints_hits(session)
        dicts = []
        for et in end_hits:
            dicts.append({'name': et[0], 'hits': et[1]})
        return jsonify(dicts)


@blueprint.route('/api/api_performance', methods=['POST'])
@secure
def api_performance():
    """
    input must be a JSON-object, with the following value: {
          'data': {
            'endpoints': ['endpoint', 'endpoint2']
          }
        }
    :return: A JSON-list for every endpoint with the following JSON-object: {
          'name': 'endpoint',
          'values': [100, 101, 102, ...]
        }
    """
    data = json.loads(request.data)['data']
    endpoints = data['endpoints']

    with session_scope() as session:
        return jsonify(get_api_performance(session, endpoints))


@blueprint.route('/api/set_rule', methods=['POST'])
@admin_secure
def set_rule():
    """
        The data from the form is validated and processed, such that the required rule is monitored
    """
    endpoint_name = request.form['name']
    value = int(request.form['value'])
    with session_scope() as session:
        set_endpoint_rule(session, endpoint_name, value)
    return 'OK'


@blueprint.route('/api/endpoint_info/<endpoint_id>')
@secure
def endpoint_info(endpoint_id):
    """
    :return: A JSON-object with endpoint details, such as:
        - color: hashed color of the endpoint,
        - endpoint: endpoint-name
        - id: endpoint id
        - methods: HTTP methods (encoded as a JSON-list)
        - monitor-level: monitor level for this endpoint
        - rules: all rules for this endpoint (encoded as a JSON-list)
        - total_hits: number of hits
        - url: link to this endpoint
    """
    with session_scope() as session:
        return jsonify(get_endpoint_details(session, endpoint_id))


@blueprint.route('api/endpoint_status_code_distribution/<endpoint_id>')
@secure
def endpoint_status_code_distribution(endpoint_id):
    with session_scope() as session:
        return jsonify(get_status_code_distribution(session, endpoint_id))


@blueprint.route('api/endpoint_status_code_summary/<endpoint_id>')
@secure
def endpoint_status_code_summary(endpoint_id):
    with session_scope() as session:
        result = {
            'distribution': get_status_code_distribution(session, endpoint_id),
            'error_requests': [
                row2dict(row) for row in get_error_requests(session, endpoint_id)
            ],
        }
        return jsonify(result)


@blueprint.route('api/endpoint_versions/<endpoint_id>', methods=['POST'])
@secure
def endpoint_versions(endpoint_id):
    with session_scope() as session:
        data = json.loads(request.data)['data']
        versions = data['versions']
        return jsonify(get_endpoint_versions(session, endpoint_id, versions))


@blueprint.route('/api/endpoint_users/<endpoint_id>', methods=['POST'])
@secure
def endpoint_users(endpoint_id):
    with session_scope() as session:
        data = json.loads(request.data)['data']
        users = data['users']
        return jsonify(get_endpoint_users(session, endpoint_id, users))



from datetime import datetime

import pytest


@pytest.mark.parametrize('request_1__time_requested', [datetime(2020, 1, 1)])
@pytest.mark.parametrize('request_2__time_requested', [datetime(2020, 1, 2)])
@pytest.mark.usefixtures('request_1', 'request_2')
def test_num_requests(dashboard_user, endpoint):
    response = dashboard_user.get('dashboard/api/requests/2020-01-01/2020-01-02')

    assert response.status_code == 200

    assert response.json['days'] == ['2020-01-01', '2020-01-02']
    [data] = [row for row in response.json['data'] if row['name'] == endpoint.name]
    assert data['values'] == [1, 1]


@pytest.mark.parametrize('request_1__time_requested', [datetime(2020, 1, 1, hour=2)])
@pytest.mark.parametrize('request_2__time_requested', [datetime(2020, 1, 1, hour=3)])
@pytest.mark.usefixtures('request_1', 'request_2')
def test_hourly_load(dashboard_user, endpoint):
    response = dashboard_user.get('dashboard/api/hourly_load/2020-01-01/2020-01-01/{0}'.format(endpoint.id))

    assert response.status_code == 200

    data = response.json
    assert data['days'] == ['2020-01-01']
    for index, row in enumerate(data['data']):
        if index in [2, 3]:
            assert row == [1]
        else:
            assert row == [0]




"""
This file contains all unit tests for the endpoint-table in the database. (Corresponding to the
file: 'flask_monitoringdashboard/database/request.py')
"""
from __future__ import division  # can be removed once we leave python 2.7

import time
from datetime import datetime, timedelta

import pytest

from flask_monitoringdashboard.core.date_interval import DateInterval
from flask_monitoringdashboard.database.count import count_requests
from flask_monitoringdashboard.database.endpoint import get_avg_duration, get_endpoints
from flask_monitoringdashboard.database.request import add_request, \
    get_date_of_first_request, get_latencies_sample, create_time_based_sample_criterion
from flask_monitoringdashboard.database.versions import get_versions


def test_get_latencies_sample(session, request_1, endpoint):
    interval = DateInterval(datetime.utcnow() - timedelta(days=1), datetime.utcnow())
    requests_criterion = create_time_based_sample_criterion(interval.start_date(),
                                                            interval.end_date())
    data = get_latencies_sample(session, endpoint.id, requests_criterion, sample_size=500)
    assert data == [request_1.duration]


def test_add_request(endpoint, session):
    num_requests = len(endpoint.requests)
    add_request(
        session,
        duration=200,
        endpoint_id=endpoint.id,
        ip='127.0.0.1',
        group_by=None,
        status_code=200,
    )
    assert count_requests(session, endpoint.id) == num_requests + 1


@pytest.mark.parametrize('request_1__time_requested', [datetime(2020, 2, 3)])
def test_get_versions(session, request_1):
    for version, first_request in get_versions(session):
        if version == request_1.version_requested:
            assert first_request == request_1.time_requested
            return
    assert False, "Shouldn't reach here"


def test_get_endpoints(session, endpoint):
    endpoints = get_endpoints(session)
    assert endpoint.name in [endpoint.name for endpoint in endpoints]


@pytest.mark.parametrize('request_1__time_requested', [datetime(1970, 1, 1)])
def test_get_date_of_first_request(session, request_1):
    total_seconds = int(time.mktime(request_1.time_requested.timetuple()))
    assert get_date_of_first_request(session) == total_seconds


def test_get_avg_duration(session, request_1, request_2, endpoint):
    assert get_avg_duration(session, endpoint.id) == (request_1.duration + request_2.duration) / 2



import datetime

import numpy
from sqlalchemy import func, and_

from flask_monitoringdashboard.core.timezone import to_utc_datetime, to_local_datetime
from flask_monitoringdashboard.database import Request
from flask_monitoringdashboard.database.count_group import count_requests_per_day, get_value
from flask_monitoringdashboard.database.endpoint import get_endpoints, get_num_requests
from flask_monitoringdashboard.database.request import create_time_based_sample_criterion


def get_num_requests_data(session, start_date, end_date):
    """
    :param session: session for the database
    :param start_date: datetime object
    :param end_date: datetime object and: end_date >= start_date
    :return: a list of the number of requests for each endpoint and on which day
    """
    numdays = (end_date - start_date).days + 1
    days = [start_date + datetime.timedelta(days=i) for i in range(numdays)]

    hits = count_requests_per_day(session, days)
    endpoints = get_endpoints(session)
    data = [
        {'name': end.name, 'values': [get_value(hits_day, end.id) for hits_day in hits]}
        for end in endpoints
    ]

    return {'days': [d.strftime('%Y-%m-%d') for d in days], 'data': data}


def get_all_request_status_code_counts(session, endpoint_id):
    """
    Gets all the request status code counts.

    :param session: session for the database
    :param endpoint_id: id for the endpoint
    :return: A list of tuples in the form of `(status_code, count)`
    """
    return (
        session.query(Request.status_code, func.count(Request.status_code))
            .filter(Request.endpoint_id == endpoint_id, Request.status_code.isnot(None))
            .group_by(Request.status_code)
            .all()
    )


def get_status_code_distribution(session, endpoint_id):
    """
    Gets the distribution of status codes returned by the given endpoint.

    :param session: session for the database
    :param endpoint_id: id for the endpoint
    :return: A dict where the key is the status code and the value is the fraction of requests
    that returned the status
    code. Example: a return value of `{ 200: 0.92, 404: 0.08 }` means that status code 200 was
    returned on 92% of the
    requests. 8% of the requests returned a 404 status code.
    """
    status_code_counts = get_all_request_status_code_counts(session, endpoint_id)
    total_count = sum(frequency for (_, frequency) in status_code_counts)
    return {status_code: frequency / total_count for (status_code, frequency) in status_code_counts}


def get_status_code_frequencies(session, endpoint_id, *criterion):
    """
    Gets the frequencies of each status code.


    :param session: session for the database
    :param endpoint_id: id for the endpoint
    :param criterion: Optional criteria used to file the requests.
    :return: A dict where the key is the status code and the value is the fraction of requests that returned the status
    code. Example: a return value of `{ 200: 105, 404: 3 }` means that status code 200 was returned 105 times and
    404 was returned 3 times.
    """
    status_code_counts = session.query(Request.status_code, func.count(Request.status_code)) \
        .filter(Request.endpoint_id == endpoint_id, Request.status_code.isnot(None), *criterion) \
        .group_by(Request.status_code).all()

    return dict(status_code_counts)


def get_error_requests(session, endpoint_id, *criterion):
    """
    Gets all requests that did not return a 200 status code.

    :param session: session for the database
    :param endpoint_id: ID of the endpoint to be queried
    :param criterion: Optional criteria used to file the requests.
    :return:
    """

    criteria = and_(
        Request.endpoint_id == endpoint_id,
        Request.status_code.isnot(None),
        Request.status_code >= 400,
        Request.status_code <= 599,
    )
    return session.query(Request).filter(criteria, *criterion).all()


def get_status_code_frequencies_in_interval(session, endpoint_id, criterion):
    return get_status_code_frequencies(session, endpoint_id, *criterion)


def get_hourly_load(session, endpoint_id, start_date, end_date):
    """
    :param session: session for the database
    :param endpoint_id: id for the endpoint
    :param start_date: datetime object
    :param end_date: datetime object and: end_date >= start_date
    :return:
    """
    numdays = (end_date - start_date).days + 1

    # list of hours: 0:00 - 23:00
    hours = ['0{}:00'.format(h) for h in range(0, 10)] + ['{}:00'.format(h) for h in range(10, 24)]
    heatmap_data = numpy.zeros((len(hours), numdays))

    start_datetime = to_utc_datetime(
        datetime.datetime.combine(start_date, datetime.time(0, 0, 0, 0))
    )
    end_datetime = to_utc_datetime(datetime.datetime.combine(end_date, datetime.time(23, 59, 59)))

    for time, count in get_num_requests(session, endpoint_id, start_datetime, end_datetime):
        parsed_time = datetime.datetime.strptime(time, '%Y-%m-%d %H:%M:%S')
        day_index = (parsed_time - start_datetime).days
        hour_index = int(to_local_datetime(parsed_time).strftime('%H'))
        heatmap_data[hour_index][day_index] = count
    return {
        'days': [
            (start_date + datetime.timedelta(days=i)).strftime('%Y-%m-%d') for i in range(numdays)
        ],
        "data": heatmap_data.tolist(),
    }



import pytest

from flask_monitoringdashboard.core.timezone import to_local_datetime


@pytest.mark.usefixtures('stack_line', 'stack_line_2')
def test_num_profiled(dashboard_user, endpoint):
    response = dashboard_user.get('dashboard/api/num_profiled/{0}'.format(endpoint.id))
    assert response.status_code == 200

    data = response.json
    assert data == 2


@pytest.mark.parametrize('request_1__group_by', ['group_by'])
@pytest.mark.parametrize('offset,per_page', [[0, 10]])
def test_profiler_table(dashboard_user, stack_line, request_1, endpoint, offset, per_page):
    response = dashboard_user.get(
        'dashboard/api/profiler_table/{0}/{1}/{2}'.format(endpoint.id, offset, per_page),
    )
    assert response.status_code == 200

    [data] = response.json
    assert data['duration'] == str(request_1.duration)
    assert data['endpoint_id'] == str(endpoint.id)
    assert data['group_by'] == request_1.group_by
    assert data['id'] == str(request_1.id)
    assert data['ip'] == request_1.ip
    assert data['status_code'] == str(request_1.status_code)
    assert data['time_requested'] == str(to_local_datetime(request_1.time_requested))
    assert data['version_requested'] == request_1.version_requested

    assert len(data['stack_lines']) == 1
    assert data['stack_lines'][0]['code']['code'] == stack_line.code.code
    assert data['stack_lines'][0]['code']['filename'] == stack_line.code.filename
    assert data['stack_lines'][0]['code']['function_name'] == stack_line.code.function_name
    assert data['stack_lines'][0]['code']['line_number'] == str(stack_line.code.line_number)


def test_grouped_profiler(dashboard_user, stack_line, endpoint):
    response = dashboard_user.get('dashboard/api/grouped_profiler/{0}'.format(endpoint.id))
    assert response.status_code == 200

    [data] = response.json
    assert data['code'] == stack_line.code.code
    assert data['duration'] == stack_line.duration
    assert data['hits'] == 1
    assert data['indent'] == stack_line.indent
    assert data['std'] == 0
    assert data['total_hits'] == 1



"""Contains custom commands for the Flask-MonitoringDashboard
For a list of all commands, open a terminal and type:

>>> flask fmd --help
"""

import click
from flask.cli import with_appcontext


@click.group()
def fmd():
    pass


@fmd.command()
@with_appcontext
def init_db():
    # Importing the database package is enough
    import flask_monitoringdashboard.database

    print('Flask-MonitoringDashboard database has been created')



from numpy import median

from flask_monitoringdashboard.database import Request


def get_data_grouped(session, column, func, *where):
    """ Return the data for a specific endpoint. The result is grouped on column
    :param session: session for the database
    :param column: the column that is used for grouping
    :param func: the function to reduce the data
    :param where: additional where clause
    """
    result = session.query(column, Request.duration).filter(*where).order_by(column).all()
    # result is now a list of tuples per request.
    return group_result(result, func)


def group_result(result, func):
    """
    :param result: A list of rows from the database: e.g. [(key, data1), (key, data2)]
    :param func: the function to reduce the data e.g. func=median
    :return: the data that is reduced. e.g. [(key, (data1+data2)/2)]
    """
    data = {}
    for key, value in result:
        if key in data.keys():
            data[key].append(value)
        else:
            data[key] = [value]
    for key in data:
        data[key] = func(data[key])
    return data.items()


def group_result_endpoint(result, func):
    """
    :param result: A list of rows from the database: e.g. [(key, data1), (key, data2)]
    :param func: the function to reduce the data e.g. func=median
    :return: the data that is reduced. e.g. [(key, (data1+data2)/2)]
    """
    data = {}
    for key, value in result:
        if key.endpoint.name in data.keys():
            data[key.endpoint.name].append(value)
        else:
            data[key.endpoint.name] = [value]
    for key in data:
        data[key] = func(data[key])
    return data.items()


def get_endpoint_data_grouped(session, func, *where):
    """
    :param session: session for the database
    :param func: the function to reduce the data
    :param where: additional where clause
    """
    return get_data_grouped(session, Request.endpoint_id, func, *where)


def get_version_data_grouped(session, func, *where):
    """
    :param session: session for the database
    :param func: the function to reduce the data
    :param where: additional where clause
    """
    return get_data_grouped(session, Request.version_requested, func, *where)


def get_user_data_grouped(session, func, *where):
    """
    :param session: session for the database
    :param func: the function to reduce the data
    :param where: additional where clause
    """
    return get_data_grouped(session, Request.group_by, func, *where)


def get_two_columns_grouped(session, column, *where):
    """
    :param session: session for the database
    :param column: column that is used for the grouping (together with the Request.version)
    :param where: additional where clause
    """
    result = (
        session.query(column, Request.version_requested, Request.duration).filter(*where).all()
    )
    result = [((g, v), t) for g, v, t in result]
    return group_result(result, median)



import threading

from flask_monitoringdashboard.core.get_ip import get_ip
from flask_monitoringdashboard.core.group_by import get_group_by
from flask_monitoringdashboard.core.profiler.base_profiler import BaseProfiler
from flask_monitoringdashboard.core.profiler.outlier_profiler import OutlierProfiler
from flask_monitoringdashboard.core.profiler.performance_profiler import PerformanceProfiler
from flask_monitoringdashboard.core.profiler.stacktrace_profiler import StacktraceProfiler


def start_thread_last_requested(endpoint):
    """
    Starts a thread that updates the last_requested time in the database.
    :param endpoint: Endpoint object
    """
    BaseProfiler(endpoint).start()


def start_performance_thread(endpoint, duration, status_code):
    """
    Starts a thread that updates performance, utilization and last_requested in the database.
    :param endpoint: Endpoint object
    :param duration: duration of the request
    :param status_code: HTTP status code of the request
    """
    group_by = get_group_by()
    PerformanceProfiler(endpoint, get_ip(), duration, group_by, status_code).start()


def start_profiler_thread(endpoint):
    """ Starts a thread that profiles the main thread. """
    current_thread = threading.current_thread().ident
    group_by = get_group_by()
    thread = StacktraceProfiler(current_thread, endpoint, get_ip(), group_by)
    thread.start()
    return thread


def start_outlier_thread(endpoint):
    """ Starts a thread that collects outliers."""
    current_thread = threading.current_thread().ident
    group_by = get_group_by()
    thread = OutlierProfiler(current_thread, endpoint, get_ip(), group_by)
    thread.start()
    return thread


def start_profiler_and_outlier_thread(endpoint):
    """ Starts two threads: PerformanceProfiler and StacktraceProfiler.  """
    current_thread = threading.current_thread().ident
    ip = get_ip()
    group_by = get_group_by()
    outlier = OutlierProfiler(current_thread, endpoint, ip, group_by)
    thread = StacktraceProfiler(current_thread, endpoint, ip, group_by, outlier)
    thread.start()
    outlier.start()
    return thread



import pytest
from flask_monitoringdashboard.core.rules import get_rules


@pytest.mark.usefixtures('request_context')
def test_rules(endpoint):
    assert len(get_rules()) == 1
    assert len(get_rules(endpoint.name)) == 1
    assert get_rules('unknown') == []



import uuid
from datetime import datetime, timedelta
from random import random

import factory
from pytest_factoryboy import register, LazyFixture

from flask_monitoringdashboard.core.profiler.util import PathHash
from flask_monitoringdashboard.core.profiler.util.grouped_stack_line import GroupedStackLine
from flask_monitoringdashboard.core.profiler.util.string_hash import StringHash
from flask_monitoringdashboard.database import (
    Endpoint,
    Request,
    Outlier,
    CodeLine,
    StackLine,
    CustomGraph,
    CustomGraphData,
    User,
)
from tests.fixtures.database import ModelFactory


class UserFactory(ModelFactory):
    class Meta:
        model = User
    username = factory.LazyFunction(lambda: str(uuid.uuid4()))
    password_hash = factory.LazyFunction(lambda: str(uuid.uuid4()))
    is_admin = True

    @classmethod
    def _create(cls, model_class, password_hash=None, *args, **kwargs):
        """Override _create, because we set the password differently"""
        instance = model_class(**kwargs)
        instance.password = password_hash  # store the original password
        instance.set_password(password=password_hash)
        session = cls._meta.sqlalchemy_session
        session.add(instance)
        session.commit()
        return instance


class EndpointFactory(ModelFactory):
    class Meta:
        model = Endpoint

    name = factory.LazyFunction(lambda: str(uuid.uuid4()))
    monitor_level = 1
    time_added = factory.LazyFunction(lambda: datetime.utcnow() - timedelta(days=1))
    version_added = '1.0'
    last_requested = factory.LazyFunction(datetime.utcnow)


class RequestFactory(ModelFactory):
    class Meta:
        model = Request

    endpoint = factory.SubFactory(EndpointFactory)
    duration = factory.LazyFunction(lambda: random() * 5000)
    time_requested = factory.LazyFunction(datetime.utcnow)
    version_requested = factory.LazyFunction(lambda: str(uuid.uuid4()))
    group_by = None
    ip = factory.Faker('ipv4_private')
    status_code = 200


class OutlierFactory(ModelFactory):
    class Meta:
        model = Outlier

    request = None
    cpu_percent = None
    memory = None
    request_environment = None
    request_header = None
    request_url = None
    stacktrace = None


class CodeLineFactory(ModelFactory):
    class Meta:
        model = CodeLine

    filename = 'abc.py'
    line_number = factory.LazyFunction(lambda: int(random() * 100))
    function_name = 'f'
    code = 'a=b'


class StackLineFactory(ModelFactory):
    class Meta:
        model = StackLine

    request = None
    code = factory.SubFactory(CodeLineFactory)
    position = 0
    indent = 0
    duration = 12


class CustomGraphFactory(ModelFactory):
    class Meta:
        model = CustomGraph

    title = factory.Faker('name')
    time_added = factory.LazyFunction(datetime.utcnow)
    version_added = factory.LazyFunction(lambda: str(uuid.uuid4()))


class CustomGraphDataFactory(ModelFactory):
    class Meta:
        model = CustomGraphData

    graph = factory.SubFactory(CustomGraphFactory)
    time = factory.LazyFunction(datetime.utcnow)
    value = factory.LazyFunction(random)


class GroupedStackLineFactory(factory.Factory):
    class Meta:
        model = GroupedStackLine

    indent = 0
    code = 'code'
    values = [10, 10, 40]
    total_sum = 100
    total_hits = 6


class StringHashFactory(factory.Factory):
    class Meta:
        model = StringHash


class PathHashFactory(factory.Factory):
    class Meta:
        model = PathHash

    _string_hash = factory.SubFactory(StringHashFactory)

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        obj = model_class()
        obj._string_hash = kwargs.get('_string_hash')
        return obj


register(UserFactory, 'user')
register(UserFactory, 'another_user')
register(EndpointFactory, 'endpoint')
register(RequestFactory, 'request_1')  # unfortunately, we can't use fixture name: 'request'
register(RequestFactory, 'request_2')
register(OutlierFactory, 'outlier_1', request=LazyFixture('request_1'))
register(OutlierFactory, 'outlier_2', request=LazyFixture('request_2'))
register(CodeLineFactory, 'code_line')
register(StackLineFactory, 'stack_line', request=LazyFixture('request_1'))
register(StackLineFactory, 'stack_line_2', request=LazyFixture('request_2'), indent=1)
register(CustomGraphFactory, 'custom_graph')
register(CustomGraphDataFactory, 'custom_graph_data')

register(GroupedStackLineFactory, 'grouped_stack_line')
register(StringHashFactory, 'string_hash')
register(PathHashFactory, 'path_hash')



import numpy as np
from flask import url_for
from werkzeug.routing import BuildError

from flask_monitoringdashboard import config
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.rules import get_rules
from flask_monitoringdashboard.core.timezone import to_local_datetime
from flask_monitoringdashboard.database.count import count_requests, count_total_requests
from flask_monitoringdashboard.database.endpoint import get_endpoint_by_id
from flask_monitoringdashboard.database.request import (
    get_date_of_first_request,
    get_date_of_first_request_version,
)


def get_endpoint_details(session, endpoint_id):
    """
    Returns details about an endpoint.
    :param session: session for the database
    :param endpoint_id: id of the endpoint
    :return dictionary
    """
    endpoint = get_endpoint_by_id(session, endpoint_id)
    endpoint.time_added = to_local_datetime(endpoint.time_added)
    flask_rule = get_rules(endpoint.name)
    methods = [list(rule.methods) for rule in flask_rule]
    methods = sum(methods, [])  # flatten list
    return {
        'id': endpoint_id,
        'color': get_color(endpoint.name),
        'methods': list(dict.fromkeys(methods)),
        'endpoint': endpoint.name,
        'rules': [r.rule for r in get_rules(endpoint.name)],
        'monitor-level': endpoint.monitor_level,
        'url': get_url(endpoint.name),
        'total_hits': count_requests(session, endpoint.id),
    }


def get_details(session):
    """
    Returns details about the deployment.
    :param session: session for the database
    :return dictionary
    """
    import json
    from flask_monitoringdashboard import loc

    with open(loc() + 'constants.json', 'r') as f:
        constants = json.load(f)

    return {
        'link': config.link,
        'dashboard-version': constants['version'],
        'config-version': config.version,
        'first-request': get_date_of_first_request(session),
        'first-request-version': get_date_of_first_request_version(session, config.version),
        'total-requests': count_total_requests(session),
    }


def get_url(end):
    """
    Returns the URL if possible.
    URL's that require additional arguments, like /static/<file> cannot be retrieved.
    :param end: the endpoint for the url.
    :return: the url_for(end) or None,
    """
    try:
        return url_for(end)
    except BuildError:
        return None


def simplify(values, n=5):
    """
    Simplify a list of values. It returns a list that is representative for the input
    :param values: list of values
    :param n: length of the returned list
    :return: list with n values: min, q1, median, q3, max
    """
    if len(values) <= n:
        return values
    return [np.percentile(values, i * 100 // (n - 1)) for i in range(n)]



from datetime import timedelta

from sqlalchemy.orm.exc import NoResultFound

from flask_monitoringdashboard.database import CustomGraph, CustomGraphData, row2dict


def get_graph_id_from_name(session, name):
    """
    :param session: session for the database
    :param name: name of the graph (must be unique)
    :return: the graph_id corresponding to the name. If the name does not exists in the db,
             a new graph is added to the database.
    """
    try:
        result = session.query(CustomGraph).filter(CustomGraph.title == name).one()
    except NoResultFound:
        result = CustomGraph(title=name)
        session.add(result)
        session.flush()
    session.expunge(result)
    return result.graph_id


def add_value(session, graph_id, value):
    data = CustomGraphData(graph_id=graph_id, value=value)
    session.add(data)


def get_graphs(session):
    return session.query(CustomGraph).all()


def get_graph_data(session, graph_id, start_date, end_date):
    """
    :param session: session for the database
    :param graph_id: id to filter on
    :param start_date: Datetime object that denotes the beginning of the interval
    :param end_date: Datetime object that denotes the end of the interval
    :return: A list with values retrieved from the database
    """
    return [
        row2dict(row)
        for row in session.query(CustomGraphData)
        .filter(
            CustomGraphData.graph_id == graph_id,
            CustomGraphData.time >= start_date,
            CustomGraphData.time < end_date + timedelta(days=1),
        )
        .all()
    ]



import datetime

from sqlalchemy import func

from flask_monitoringdashboard.core.timezone import to_utc_datetime
from flask_monitoringdashboard.database import Request


def count_rows_group(session, column, *criterion):
    """
    Count the number of rows of a specified column
    :param session: session for the database
    :param column: column to count
    :param criterion: where-clause of the query
    :return list with the number of rows per endpoint
    """
    return (
        session.query(Request.endpoint_id, func.count(column))
        .filter(*criterion)
        .group_by(Request.endpoint_id)
        .all()
    )


def get_value(tuples, name, default=0):
    """
    :param tuples: must be structured as: [(a, b), (c, d), ..]
    :param name: name to filter on, e.g.: if name == a, it returns b
    :param default: returned if the name was not found in the list
    :return value corresponding to the name in the list.
    """
    for key, value in tuples:
        if key == name:
            return value
    return default


def count_requests_group(session, *where):
    """
    Return the number of hits for all endpoints (possible with more filter arguments).
    :param session: session for the database
    :param where: additional arguments
    """
    return count_rows_group(session, Request.id, *where)


def count_requests_per_day(session, list_of_days):
    """ Return the number of hits for all endpoints per day.
    :param session: session for the database
    :param list_of_days: list with datetime.datetime objects. """
    result = []
    for day in list_of_days:
        dt_begin = to_utc_datetime(datetime.datetime.combine(day, datetime.time(0, 0, 0)))
        dt_end = dt_begin + datetime.timedelta(days=1)

        result.append(
            count_rows_group(
                session,
                Request.id,
                Request.time_requested >= dt_begin,
                Request.time_requested < dt_end,
            )
        )
    return result



import pytest

from flask_monitoringdashboard.database import Endpoint, Request


@pytest.mark.parametrize('endpoint__monitor_level', [3])
def test_overview(dashboard_user, request_1, request_2, endpoint, session):
    response = dashboard_user.get('dashboard/api/overview')
    assert response.status_code == 200

    [data] = [row for row in response.json if row['id'] == endpoint.id]

    assert data['hits-overall'] == 2
    assert data['hits-today'] == 2
    assert data['hits-today-errors'] == 0
    assert data['hits-week'] == 2
    assert data['hits-week-errors'] == 0

    assert data['last-accessed'] == endpoint.last_requested.strftime("%a, %d %b %Y %H:%M:%S GMT")

    assert data['median-overall'] == (request_1.duration + request_2.duration) / 2
    assert data['median-today'] == (request_1.duration + request_2.duration) / 2
    assert data['median-week'] == (request_1.duration + request_2.duration) / 2

    assert data['monitor'] == 3
    assert data['name'] == endpoint.name
    assert data['blueprint'] == endpoint.name


@pytest.mark.parametrize('request_1__group_by', ['42'])
@pytest.mark.parametrize('request_2__group_by', ['something else'])
@pytest.mark.usefixtures('request_1', 'request_2')
def test_users(dashboard_user, endpoint, session):
    response = dashboard_user.get('dashboard/api/users/{0}'.format(endpoint.id))
    assert response.status_code == 200

    data = response.json
    row1, row2 = data

    assert row1['hits'] == 1
    assert row1['user'] == '42' or 'something else'
    assert row2['hits'] == 1
    assert row2['user'] == '42' or 'something else'


@pytest.mark.parametrize('request_1__ip', ['42'])
@pytest.mark.parametrize('request_2__ip', ['something else'])
@pytest.mark.usefixtures('request_1', 'request_2')
def test_ips(dashboard_user, endpoint, session):
    response = dashboard_user.get('dashboard/api/ip/{0}'.format(endpoint.id))
    assert response.status_code == 200

    data = response.json
    row1, row2 = data

    assert row1['hits'] == 1
    assert row1['ip'] == '42' or 'something else'
    assert row2['hits'] == 1
    assert row2['ip'] == '42' or 'something else'


def test_endpoints(dashboard_user, endpoint, session):
    response = dashboard_user.get('dashboard/api/endpoints')
    assert response.status_code == 200

    assert len(response.json) == session.query(Endpoint).count()
    [data] = [row for row in response.json if row['id'] == str(endpoint.id)]

    assert data['last_requested'] == str(endpoint.last_requested)
    assert data['monitor_level'] == str(endpoint.monitor_level)
    assert data['name'] == endpoint.name
    assert data['time_added'] == str(endpoint.time_added)
    assert data['version_added'] == endpoint.version_added


@pytest.mark.usefixtures('request_1', 'request_2')
def test_endpoint_hits(dashboard_user, endpoint, session):
    response = dashboard_user.get('dashboard/api/endpoints_hits')
    assert response.status_code == 200

    total_hits = sum(row['hits'] for row in response.json)
    assert total_hits == session.query(Request).count()

    [data] = [row for row in response.json if row['name'] == endpoint.name]
    assert data['hits'] == 2


def test_api_performance_get(dashboard_user):
    """GET is not allowed. It should return the overview page."""
    response = dashboard_user.get('dashboard/api/api_performance')
    assert not response.is_json


def test_api_performance_post(dashboard_user, request_1, endpoint, session):
    response = dashboard_user.post('dashboard/api/api_performance', json={
        'data': {'endpoints': [endpoint.name]}
    })

    assert response.status_code == 200
    [data] = response.json
    assert data['values'] == [request_1.duration]


def test_set_rule_get(dashboard_user):
    """GET is not allowed. It should return the overview page."""
    response = dashboard_user.get('dashboard/api/set_rule')
    assert not response.is_json


@pytest.mark.parametrize('user__is_admin', [False])
def test_set_rule_post_guest_not_allowed(dashboard_user, endpoint):
    """Guest is redirected to the login page."""
    response = dashboard_user.post('dashboard/api/set_rule', data={
        'name': endpoint.name,
        'value': 3,
    })
    assert response.status_code == 302


@pytest.mark.parametrize('endpoint__monitor_level', [1])
def test_set_rule_post(dashboard_user, endpoint, session):
    response = dashboard_user.post('dashboard/api/set_rule', data={
        'name': endpoint.name,
        'value': 3,
    })

    assert response.status_code == 200
    assert response.data == b'OK'
    endpoint = session.query(Endpoint).get(endpoint.id)  # reload the endpoint
    assert endpoint.monitor_level == 3


@pytest.mark.usefixtures('request_1')
def test_endpoint_info(dashboard_user, endpoint):
    response = dashboard_user.get('dashboard/api/endpoint_info/{0}'.format(endpoint.id))
    assert response.status_code == 200
    data = response.json

    assert data['endpoint'] == endpoint.name
    assert data['monitor-level'] == endpoint.monitor_level
    assert data['total_hits'] == 1
    assert data['rules'] == ['/']
    assert data['url'] == '/'


@pytest.mark.parametrize('request_1__status_code', [404])
@pytest.mark.parametrize('request_2__status_code', [200])
@pytest.mark.usefixtures('request_1', 'request_2')
def test_endpoint_status_code_distribution(dashboard_user, endpoint):
    response = dashboard_user.get('dashboard/api/endpoint_status_code_distribution/{0}'.format(endpoint.id))
    assert response.status_code == 200

    data = response.json
    assert data['200'] == 1 / 2
    assert data['404'] == 1 / 2


@pytest.mark.parametrize('request_1__status_code', [404])
@pytest.mark.parametrize('request_2__status_code', [200])
@pytest.mark.usefixtures('request_2')
def test_endpoint_status_code_summary(dashboard_user, request_1, endpoint):
    response = dashboard_user.get('dashboard/api/endpoint_status_code_summary/{0}'.format(endpoint.id))
    assert response.status_code == 200

    data = response.json
    assert data['distribution'] == {'200': 1 / 2, '404': 1 / 2}
    [row] = data['error_requests']

    assert row['duration'] == str(request_1.duration)
    assert row['endpoint_id'] == str(endpoint.id)
    assert row['group_by'] == str(request_1.group_by)
    assert row['id'] == str(request_1.id)
    assert row['ip'] == request_1.ip
    assert row['time_requested'] == str(request_1.time_requested)
    assert row['version_requested'] == request_1.version_requested


def test_endpoint_versions_get(dashboard_user, endpoint):
    """GET is not allowed. It should return the overview page."""
    response = dashboard_user.get('dashboard/api/endpoint_versions/{0}'.format(endpoint.id))
    assert not response.is_json


@pytest.mark.parametrize('request_1__version_requested', ['a'])
@pytest.mark.parametrize('request_2__version_requested', ['b'])
def test_endpoint_versions_post(dashboard_user, request_1, request_2, endpoint):
    response = dashboard_user.post(
        'dashboard/api/endpoint_versions/{0}'.format(endpoint.id),
        json={'data': {'versions': [request_1.version_requested, request_2.version_requested]}},
    )
    assert response.status_code == 200
    row1, row2 = response.json

    assert row1['version'] == request_1.version_requested
    assert row1['values'] == [request_1.duration]

    assert row2['version'] == request_2.version_requested
    assert row2['values'] == [request_2.duration]


def test_endpoint_users_get(dashboard_user, endpoint):
    """GET is not allowed. It should return the overview page."""
    response = dashboard_user.get('dashboard/api/endpoint_users/{0}'.format(endpoint.id))
    assert not response.is_json


@pytest.mark.parametrize('request_1__group_by', ['a'])
@pytest.mark.parametrize('request_2__group_by', ['b'])
def test_endpoint_users_post(dashboard_user, request_1, request_2, endpoint):
    response = dashboard_user.post(
        'dashboard/api/endpoint_users/{0}'.format(endpoint.id),
        json={'data': {'users': [request_1.group_by, request_2.group_by]}},
    )
    assert response.status_code == 200
    row1, row2 = response.json

    assert row1['user'] == request_1.group_by
    assert row1['values'] == [request_1.duration]

    assert row2['user'] == request_2.group_by
    assert row2['values'] == [request_2.duration]


from flask_monitoringdashboard.database.stack_line import (
    add_stack_line,
    get_profiled_requests,
    get_grouped_profiled_requests,
)

from flask_monitoringdashboard.database import StackLine, Request


def test_add_stackline(session, request_1):
    assert session.query(StackLine).filter(StackLine.request_id == request_1.id).one_or_none() is None
    add_stack_line(session, request_id=request_1.id, position=0, indent=1, duration=1, code_line="code")
    session.commit()
    assert session.query(StackLine).filter(StackLine.request_id == request_1.id).one()


def test_get_profiled_requests(session, endpoint, request_1):
    assert not get_profiled_requests(session, endpoint_id=endpoint.id, offset=0, per_page=10)
    add_stack_line(session, request_id=request_1.id, position=0, indent=1, duration=1, code_line="code")
    session.commit()
    assert get_profiled_requests(session, endpoint_id=endpoint.id, offset=0, per_page=10)


def test_get_grouped_profiled_requests(session, request_1, endpoint):
    assert not get_grouped_profiled_requests(session, endpoint_id=endpoint.id)
    add_stack_line(session, request_id=request_1.id, position=0, indent=1, duration=1, code_line="code")
    session.commit()
    assert get_grouped_profiled_requests(session, endpoint_id=endpoint.id)


from flask_monitoringdashboard.database.data_grouped import (
    get_endpoint_data_grouped,
    get_version_data_grouped,
)


def test_get_endpoint_data_grouped(session, request_1):
    data = get_endpoint_data_grouped(session, lambda x: x)
    for key, value in data:
        if key == request_1.endpoint_id:
            assert value == [request_1.duration]
            return
    assert False, "Shouldn't reach here."


def test_get_version_data_grouped(session, request_1):
    data = get_version_data_grouped(session, lambda x: x)
    for key, value in data:
        if key == request_1.version_requested:
            assert value == [request_1.duration]
            return
    assert False, "Shouldn't reach here."



from flask_monitoringdashboard.database import Request


def test_deployment(dashboard_user, session, config):
    response = dashboard_user.get('dashboard/api/deploy_details')
    assert response.status_code == 200

    data = response.json
    assert data['config-version'] == config.version
    assert data['link'] == 'dashboard'
    assert data['total-requests'] == session.query(Request).count()


def test_deployment_config(dashboard_user, config):
    response = dashboard_user.get('dashboard/api/deploy_config')
    assert response.status_code == 200

    data = response.json
    assert data['database_name'] == config.database_name
    assert data['outlier_detection_constant'] == config.outlier_detection_constant
    assert data['timezone'] == str(config.timezone)
    assert data['colors'] == config.colors



import pytest


def test_stringhash(string_hash):
    assert string_hash.hash('abc') == 0
    assert string_hash.hash('def') == 1
    assert string_hash.hash('abc') == 0


def test_unhash(string_hash):
    assert string_hash.unhash(string_hash.hash('abc')) == 'abc'

    with pytest.raises(ValueError):
        string_hash.unhash('unkown')



import threading
import time

from flask import request
import pytest
from werkzeug.routing import Rule

from flask_monitoringdashboard.core.cache import memory_cache, init_cache
from flask_monitoringdashboard.core.profiler import (
    start_thread_last_requested,
    start_performance_thread,
    start_profiler_and_outlier_thread,
    start_outlier_thread,
)


def wait_until_threads_finished(num_threads):
    while threading.active_count() > num_threads:
        time.sleep(0.01)


@pytest.mark.usefixtures('request_context')
def test_start_thread_last_requested(endpoint, config):
    config.app.url_map.add(Rule('/', endpoint=endpoint.name))
    init_cache()
    num_threads = threading.active_count()
    start_thread_last_requested(endpoint)
    wait_until_threads_finished(num_threads)

    assert memory_cache.get(endpoint.name).last_requested


@pytest.mark.usefixtures('request_context')
def test_start_performance_thread(endpoint, config):
    config.app.url_map.add(Rule('/', endpoint=endpoint.name))
    init_cache()
    request.environ['REMOTE_ADDR'] = '127.0.0.1'
    num_threads = threading.active_count()
    start_performance_thread(endpoint, 1234, 200)
    assert threading.active_count() == num_threads + 1
    wait_until_threads_finished(num_threads)

    assert memory_cache.get(endpoint.name).average_duration > 0


@pytest.mark.usefixtures('request_context')
def test_start_outlier_thread(endpoint, config):
    config.app.url_map.add(Rule('/', endpoint=endpoint.name))
    init_cache()
    request.environ['REMOTE_ADDR'] = '127.0.0.1'
    num_threads = threading.active_count()
    outlier = start_outlier_thread(endpoint)
    assert threading.active_count() == num_threads + 1
    outlier.stop(duration=1, status_code=200)
    wait_until_threads_finished(num_threads)


@pytest.mark.usefixtures('request_context')
def test_start_profiler_and_outlier_thread(endpoint, config):
    config.app.url_map.add(Rule('/', endpoint=endpoint.name))
    init_cache()
    request.environ['REMOTE_ADDR'] = '127.0.0.1'
    num_threads = threading.active_count()
    thread = start_profiler_and_outlier_thread(endpoint)
    assert threading.active_count() == num_threads + 2
    thread.stop(duration=1, status_code=200)
    wait_until_threads_finished(num_threads)



"""Use this file for migrating the Database from v1.X.X to v2.X.X
Before running the script, make sure to change the OLD_DB_URL and NEW_DB_URL on lines 9 and 10.
Refer to http://docs.sqlalchemy.org/en/latest/core/engines.html on how to configure this.
"""
import datetime
from contextlib import contextmanager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, joinedload

from flask_monitoringdashboard.database import Endpoint, Request, Outlier

# OLD_DB_URL = 'dialect+driver://username:password@host:port/old_db'
# NEW_DB_URL = 'dialect+driver://username:password@host:port/new_db'
OLD_DB_URL = 'sqlite:////home/bogdan/databases/flask-dashboard.db'
NEW_DB_URL = 'sqlite:////home/bogdan/databases/flask-dashboard-2.0.db'
# OLD_DB_URL = 'mysql+pymysql://root:admin@localhost/migration1'
# NEW_DB_URL = 'mysql+pymysql://root:admin@localhost/migration2'


TABLES = ["rules", "functionCalls", "outliers", "tests", "testRun"]
DATE_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
SEARCH_REQUEST_TIME = datetime.timedelta(seconds=10)

endpoint_dict = {}
outlier_dict = {}
tests_dict = {}


def create_new_db(db_url):
    from flask_monitoringdashboard.database import Base

    engine = create_engine(db_url)
    Base.metadata.drop_all(engine)
    Base.metadata.create_all(engine)
    Base.metadata.bind = engine
    global DBSession
    DBSession = sessionmaker(bind=engine)


def get_connection(db_url):
    engine = create_engine(db_url)
    connection = engine.connect()
    return connection


@contextmanager
def session_scope():
    """When accessing the database, use the following syntax:
    >>> with session_scope() as session:
    >>>     session.query(...)

    :return: the session for accessing the database
    """
    session = DBSession()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def get_session(db_url):
    """This creates the new database and returns the session scope."""
    from flask_monitoringdashboard import config

    config.database_name = db_url

    import flask_monitoringdashboard.database

    return flask_monitoringdashboard.database.session_scope()


def parse(date_string):
    if not date_string:
        return None
    if isinstance(date_string, str):
        return datetime.datetime.strptime(date_string, DATE_FORMAT)
    return date_string


def move_rules(old_connection):
    rules = old_connection.execute("select * from {}".format(TABLES[0]))
    endpoints = []
    with session_scope() as session:
        for rule in rules:
            end = Endpoint(
                name=rule['endpoint'],
                monitor_level=rule['monitor'],
                time_added=parse(rule['time_added']),
                version_added=rule['version_added'],
                last_requested=parse(rule['last_accessed']),
            )
            endpoints.append(end)
        session.bulk_save_objects(endpoints)


def populate_endpoint_dict(session):
    global endpoint_dict
    endpoints = session.query(Endpoint).all()
    for endpoint in endpoints:
        endpoint_dict[endpoint.name] = endpoint.id


def move_function_calls(old_connection):
    function_calls = old_connection.execute("select * from {}".format(TABLES[1]))
    requests = []
    with session_scope() as session:
        populate_endpoint_dict(session)
        for fc in function_calls:
            request = Request(
                endpoint_id=endpoint_dict[fc['endpoint']],
                duration=fc['execution_time'],
                time_requested=parse(fc['time']),
                version_requested=fc['version'],
                group_by=fc['group_by'],
                ip=fc['ip'],
            )
            requests.append(request)
        session.bulk_save_objects(requests)


def get_request_id(requests, time, execution_time, start_index):
    for index, r in enumerate(requests):
        if index >= start_index:
            if (
                abs(r.time_requested - parse(time)) < SEARCH_REQUEST_TIME
                and r.duration == execution_time
            ):
                return r.id, index
    return None, start_index


def populate_outlier_dict(connection, session):
    global outlier_dict
    outliers = connection.execute("select * from {}".format(TABLES[2]))
    requests = session.query(Request).options(joinedload(Request.endpoint)).all()
    index = 0
    for outlier in outliers:
        req_id, index = get_request_id(
            requests, outlier['time'], outlier['execution_time'], start_index=index
        )
        outlier_dict[outlier['id']] = req_id


def move_outliers(old_connection):
    global outlier_dict
    old_outliers = old_connection.execute("select * from {}".format(TABLES[2]))
    outliers = []
    with session_scope() as session:
        populate_outlier_dict(old_connection, session)
        for o in old_outliers:
            outlier = Outlier(
                request_id=outlier_dict[o['id']],
                request_header=o['request_headers'],
                request_environment=o['request_environment'],
                request_url=o['request_url'],
                cpu_percent=o['cpu_percent'],
                memory=o['memory'],
                stacktrace=o['stacktrace'],
            )
            outliers.append(outlier)
        session.bulk_save_objects(outliers)


def main():
    create_new_db(NEW_DB_URL)
    old_connection = get_connection(OLD_DB_URL)
    import timeit

    start = timeit.default_timer()
    move_rules(old_connection)
    t1 = timeit.default_timer()
    print("Moving rules took %f seconds" % (t1 - start))
    move_function_calls(old_connection)
    t2 = timeit.default_timer()
    print("Moving functionCalls took %f seconds" % (t2 - t1))
    move_outliers(old_connection)
    t3 = timeit.default_timer()
    print("Moving outliers took %f seconds" % (t3 - t2))

    print("Total time was %f seconds" % (t3 - start))


if __name__ == "__main__":
    main()
